"""Package containing Qt widgets.

The program is organized in widgets.
Roughly one widget is stored in one module.

A module should start by importing:

 - packages that are in the Python library
 - additional packages (in this order: numpy, scipy, PyQt5)
 - relative imports

Leave an empty line between each group of imports.
Then make sure that the logging works:

    from logging import getLogger
    lg = getLogger(__name__)

In code add important messages with:

    lg.info('Important message')

and messages useful for debugging:

    lg.debug('Message for debugging')
"""
